use clap::Parser;
use helpers::command_output;
use std::str::FromStr;
use tokio::io::{self, AsyncBufReadExt, BufReader};
use warp::Filter;
use wg::Peer;

mod event;
mod helpers;
mod wg;

type WgLink = String;
type WgPeer = String;

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct CliArguments {
    #[arg(short, long)]
    addr: std::net::SocketAddr,
    #[arg(short, long)]
    link: WgLink,
}

#[derive(Debug)]
struct RejectCommandFailedToExecute;
impl warp::reject::Reject for RejectCommandFailedToExecute {}

#[derive(Debug)]
struct CommandError {
    stdout: String,
    stderr: String,
}
impl warp::reject::Reject for CommandError {}
impl From<std::process::Output> for CommandError {
    fn from(value: std::process::Output) -> Self {
        CommandError {
            stdout: std::str::from_utf8(value.stdout.as_ref()).unwrap().into(),
            stderr: std::str::from_utf8(value.stderr.as_ref()).unwrap().into(),
        }
    }
}

fn split_allowed_ips_text_line(text: &str) -> (WgPeer, Vec<cidr::IpInet>) {
    let mut string_parts = text.split(|c| c == '\t' || c == ' ');
    let peer = string_parts.next().unwrap();
    let cidrs = string_parts
        .map(|p| cidr::IpInet::from_str(p).unwrap())
        .collect();
    (peer.into(), cidrs)
}
fn split_allowed_ips_text(text: &str, peer: WgPeer) -> (WgPeer, Vec<cidr::IpInet>) {
    text.split('\n')
        .map(|p| split_allowed_ips_text_line(p))
        .filter(|l| l.0 == peer)
        .next()
        .unwrap()
}

async fn wg_add_address(
    link: WgLink,
    peer: WgPeer,
    cidr: cidr::IpInet,
) -> Result<impl warp::Reply, warp::Rejection> {
    let output = match command_output(
        tokio::process::Command::new("wg")
            .arg("show")
            .arg(&link)
            .arg("allowed-ips"),
    )
    .await
    {
        Ok(v) => v,
        Err(_) => return Err(warp::reject::custom(RejectCommandFailedToExecute)),
    };
    if !output.status.success() {
        return Err(warp::reject::custom(CommandError::from(output)));
    }
    let (peer, ips) =
        split_allowed_ips_text(std::str::from_utf8(output.stdout.as_ref()).unwrap(), peer);
    let ips_str =
        ips.iter()
            .chain([cidr].iter())
            .map(|x| x.to_string())
            .fold("".to_string(), |acc, x| {
                let mut s = String::from(acc);
                s.push_str(&x);
                s.push_str(",");
                s
            });
    let ips_str = &ips_str[0..ips_str.len() - 1];
    let mut command = tokio::process::Command::new("wg");
    command
        .arg("set")
        .arg(&link)
        .arg("peer")
        .arg(&peer)
        .arg("allowed-ips")
        .arg(ips_str);
    let output = match command_output(&mut command).await {
        Ok(v) => v,
        Err(_) => return Err(warp::reject::custom(RejectCommandFailedToExecute)),
    };

    if output.status.success() {
        Ok("")
    } else {
        Err(warp::reject::custom(CommandError::from(output)))
    }
}
async fn wg_del_address(
    link: WgLink,
    peer: WgPeer,
    cidr: cidr::IpInet,
) -> Result<impl warp::Reply, warp::Rejection> {
    let output = match tokio::process::Command::new("wg")
        .arg("show")
        .arg(&link)
        .arg("allowed-ips")
        .output()
        .await
    {
        Ok(v) => v,
        Err(_) => return Err(warp::reject::custom(RejectCommandFailedToExecute)),
    };
    if !output.status.success() {
        return Err(warp::reject::custom(CommandError::from(output)));
    }
    let (peer, ips) =
        split_allowed_ips_text(std::str::from_utf8(output.stdout.as_ref()).unwrap(), peer);
    let ips_str = ips
        .iter()
        .filter(|x| **x != cidr)
        .map(|x| x.to_string())
        .fold("".to_string(), |acc, x| {
            let mut s = String::from(acc);
            s.push_str(&x);
            s.push_str(",");
            s
        });
    let ips_str = &ips_str[0..ips_str.len() - 1];
    let mut command = tokio::process::Command::new("wg");
    command
        .arg("set")
        .arg(&link)
        .arg("peer")
        .arg(&peer)
        .arg("allowed-ips")
        .arg(ips_str);
    println!("command = {:?}", command);
    let output = match command.output().await {
        Ok(v) => v,
        Err(_) => return Err(warp::reject::custom(RejectCommandFailedToExecute)),
    };
    println!("{:?}", output);

    if output.status.success() {
        Ok("")
    } else {
        Err(warp::reject::custom(CommandError::from(output)))
    }
}

async fn send_daemon_get_public_key(iface: &WgLink) -> WgPeer {
    let mut command = tokio::process::Command::new("wg");
    command.arg("show").arg(&iface).arg("public-key");
    println!("command = {:?}", command);
    let output = command.output().await.unwrap();
    std::str::from_utf8(output.stdout.as_ref())
        .unwrap()
        .trim_end_matches("\n")
        .to_string()
}

async fn send_daemon(iface: WgLink, port: u16, peers: Vec<Peer>) -> () {
    let re_send_add = regex::Regex::new(r"^\d+:.*\n$").unwrap();
    let re_send_del = regex::Regex::new(r"^Deleted \d+:.*\n$").unwrap();
    let re_extract_ip = regex::Regex::new(r"^.*inet ((\d+\.?){4}(/\d+)?).*\n$").unwrap();
    let mut command = tokio::process::Command::new("ip");
    command.arg("monitor").arg("address").arg("dev").arg(&iface);
    command.stdout(std::process::Stdio::piped());
    command.stderr(std::process::Stdio::piped());
    log::debug!("Shell command {:?}", command);
    let mut child = command.spawn().unwrap();
    let stdout = child.stdout.take().unwrap();
    let stderr = child.stderr.take().unwrap();
    let mut reader = BufReader::new(stdout);
    let mut stderr_reader = BufReader::new(stderr);

    let mut buffer: Vec<u8> = vec![];
    let mut stderr_buffer: Vec<u8> = vec![];
    let mut enable_stdout = true;
    let mut enable_stderr = true;
    loop {
        let mut send_add = false;
        let mut send_del = false;

        let cidr_: String;
        let public_key;
        let b: String;

        if enable_stdout & enable_stderr {
            tokio::select!(
                v = reader.read_until('\n' as u8, &mut buffer) => {
                    if v.is_err() | v.is_ok_and(|v_| v_ == 0) {
                        enable_stdout = false;
                        continue;
                    }
                    b = String::from_utf8(buffer).unwrap();
                    buffer = vec![];
                },
                v = stderr_reader.read_until('\n' as u8, &mut stderr_buffer) => {
                    if v.is_err() | v.is_ok_and(|v_| v_ == 0) {
                        enable_stderr = false;
                        continue;
                    }
                    let b = String::from_utf8(stderr_buffer).unwrap();
                    stderr_buffer = vec![];
                    log::error!("Monitor produced error: {:?}", &b);
                    continue;
                },
            );
        } else if enable_stdout {
            let v = reader.read_until('\n' as u8, &mut buffer).await;
            if v.is_err() | v.is_ok_and(|v_| v_ == 0) {
                enable_stdout = false;
                continue;
            }
            b = String::from_utf8(buffer).unwrap();
            buffer = vec![];
        } else if enable_stderr {
            let v = stderr_reader
                .read_until('\n' as u8, &mut stderr_buffer)
                .await;
            if v.is_err() | v.is_ok_and(|v_| v_ == 0) {
                enable_stderr = false;
                continue;
            }
            let b = String::from_utf8(stderr_buffer).unwrap();
            stderr_buffer = vec![];
            log::error!("Monitor produced error: {:?}", &b);
            continue;
        } else {
            break;
        }

        if re_send_add.is_match(&b) {
            send_add = true;
        } else if re_send_del.is_match(&b) {
            send_del = true;
        }
        if send_add | send_del {
            cidr_ =
                urlencoding::encode(re_extract_ip.captures(&b).unwrap().get(1).unwrap().as_str())
                    .into();
            public_key = send_daemon_get_public_key(&iface).await;
        } else {
            continue;
        }
        if send_add {
            for peer in peers.iter() {
                let peer_ip = peer.allowed_ips.get(0).unwrap();
                let u = format!(
                    "http://{peer}:{port}/wireguard/{iface}/peer/{public_key}/address",
                    peer = peer_ip,
                    port = port,
                    iface = iface,
                    public_key = public_key
                );
                println!("{:?}", u);
                let request = hyper::Request::builder()
                    .method(hyper::Method::POST)
                    .uri(&u)
                    .body(hyper::Body::from(cidr_.clone()))
                    .unwrap();
                let r = hyper::Client::new().request(request).await;
                println!("{:?}", r);
            }
        } else if send_del {
            for peer in peers.iter() {
                let peer_ip = peer.allowed_ips.get(0).unwrap();
                let u = format!(
                    "http://{peer}:{port}/wireguard/{iface}/peer/{public_key}/address/{cidr}",
                    peer = peer_ip,
                    port = port,
                    iface = iface,
                    public_key = public_key,
                    cidr = &cidr_
                );
                println!("{:?}", u);
                let request = hyper::Request::builder()
                    .method(hyper::Method::DELETE)
                    .uri(&u)
                    .body(hyper::Body::empty())
                    .unwrap();
                let r = hyper::Client::new().request(request).await;
                println!("{:?}", r);
            }
        }
    }
}

const BUILD_INFO: &str = include!(concat!(env!("OUT_DIR"), "/buildinfo.txt"));

#[tokio::main]
async fn main() {
    if std::env::var("WGVIRTIPD_LOG").is_err() {
        std::env::set_var("WGVIRTIPD_LOG", "wgvirtipd=info");
    }
    env_logger::init_from_env("WGVIRTIPD_LOG");
    let args = CliArguments::parse();
    let config = wg::Wg::from_file(&args.link).await;

    let alive = warp::path!("alive").and(warp::get()).map(|| BUILD_INFO);
    let base = warp::path("wireguard")
        .and(warp::path::param::<WgLink>())
        .and(warp::path("peer"))
        .and(warp::path::param::<WgPeer>())
        .and(warp::path("address"));
    let address_add = base
        .and(warp::path::end())
        .and(warp::post())
        .and(warp::body::bytes().map(|b: bytes::Bytes| {
            cidr::IpInet::from_str(
                urlencoding::decode(std::str::from_utf8(b.as_ref()).unwrap().trim())
                    .unwrap()
                    .to_string()
                    .as_str(),
            )
            .unwrap()
        }))
        .and_then(wg_add_address);
    let address_del = base
        .and(warp::path::param::<cidr::IpInet>())
        .and(warp::path::end())
        .and(warp::delete())
        .and_then(wg_del_address);
    let routes = alive.or(address_add.or(address_del));

    log::info!("{}", BUILD_INFO);

    tokio::select!(
        _ = warp::serve(routes).run(args.addr) => {},
        _ = send_daemon(args.link, args.addr.port(), config.peers) => {},
    );
}
