use cidr::IpInet;
use serde::de::{self, Error as _, MapAccess, Visitor};
use serde::{Deserialize, Deserializer, Serialize};
use serde_with::formats::CommaSeparator;
use serde_with::StringWithSeparator;
use serde_with::{serde_as, Map, Seq};
use std::collections::HashMap;
use std::path::{Path, PathBuf};

fn de_one_many<'de, T, D>(deserializer: D) -> Result<Vec<T>, D::Error>
where
    D: Deserializer<'de>,
    T: serde::Deserialize<'de>,
{
    #[derive(Deserialize)]
    #[serde(untagged)]
    enum OneMany<T> {
        One(T),
        Vec(Vec<T>),
    }
    match OneMany::<T>::deserialize(deserializer)? {
        OneMany::One(t) => Ok(vec![t]),
        OneMany::Vec(v) => Ok(v),
    }
}

fn de_one_many_none<'de, T, D>(deserializer: D) -> Result<Option<Vec<T>>, D::Error>
where
    D: Deserializer<'de>,
    T: serde::Deserialize<'de>,
{
    #[derive(Deserialize)]
    #[serde(untagged)]
    enum OneManyNone<T> {
        None,
        One(T),
        Vec(Vec<T>),
    }
    match OneManyNone::<T>::deserialize(deserializer)? {
        OneManyNone::None => Ok(None),
        OneManyNone::One(t) => Ok(Some(vec![t])),
        OneManyNone::Vec(v) => Ok(Some(v)),
    }
}

fn de_host<'de, D>(deserializer: D) -> Result<(url::Host<String>, u16), D::Error>
where
    D: Deserializer<'de>,
{
    let v = String::deserialize(deserializer)?;
    let u = url::Url::parse(format!("a://{}", v).as_ref()).unwrap();
    let host = match u.host().unwrap() {
        url::Host::Domain(s) => url::Host::Domain(String::from(s)),
        url::Host::Ipv4(a) => url::Host::Ipv4(a),
        url::Host::Ipv6(a) => url::Host::Ipv6(a),
    };
    let port = u.port().unwrap();
    return Ok((host, port));
}

#[derive(Clone, Debug)]
#[serde_as]
#[derive(Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct Peer {
    pub public_key: String,
    pub preshared_key: Option<String>,
    #[serde(rename = "AllowedIPs")]
    #[serde_as(as = "StringWithSeparator::<CommaSeparator, IpInet>")]
    pub allowed_ips: Vec<IpInet>,
    #[serde(deserialize_with = "de_host")]
    pub endpoint: (url::Host, u16),
    pub persistent_keepalive: Option<u16>,
}

#[derive(Clone, Debug)]
#[serde_as]
#[derive(Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct Interface {
    pub private_key: String,
    pub listen_port: Option<u16>,
    pub fw_mark: Option<u32>,
    // wg-quick
    #[serde_as(as = "StringWithSeparator::<CommaSeparator, IpInet>")]
    pub address: Vec<IpInet>, // TODO address can actually be left out with normal wg set link
    #[serde(rename = "DNS")]
    pub dns: Option<url::Host>,
    #[serde(rename = "MTU")]
    pub mtu: Option<u32>,
    pub table: Option<String>,
    pub pre_up: Option<String>,
    pub post_up: Option<String>,
    pub pre_down: Option<String>,
    pub post_down: Option<String>,
    pub save_config: Option<bool>,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
enum PeerOrInterface {
    Peer(Peer),
    Interface(Interface),
}

type _Wg = Vec<PeerOrInterface>;

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Wg {
    pub interface: Interface,
    pub peers: Vec<Peer>,
}

impl Wg {
    pub fn from_str(s: &str) -> Self {
        let w: _Wg = serde_ini::de::from_str(s).unwrap();
        let mut interface: Option<Interface> = None;
        let mut peers: Vec<Peer> = vec![];
        for i in w {
            match i {
                PeerOrInterface::Interface(i) => interface = Some(i),
                PeerOrInterface::Peer(p) => peers.push(p),
            }
        }
        let interface = interface.unwrap();
        Self { interface, peers }
    }

    pub async fn from_file(path: impl AsRef<Path>) -> Self {
        let mut p: &Path = path.as_ref();
        let mut new_p;
        if p.parent() == Some(Path::new("")) {
            new_p = PathBuf::new();
            new_p.push("/etc/wireguard");
            new_p.push(format!("{}.conf", p.to_str().unwrap()));
            p = &new_p;
        }
        let s = tokio::fs::read_to_string(p).await.unwrap();
        Self::from_str(&s)
    }
}
